"""
Tool Name:     Exploratory Regression
Source Name:   ExploratoryRegression.py
Version:       ArcGIS 10.0
Author:        Environmental Systems Research Institute Inc.
"""

################ Imports ####################
import sys as SYS
import copy as COPY
import os as OS
import locale as LOCALE
import numpy as NUM
import math as MATH
import numpy.linalg as LA
import arcpy as ARCPY
import arcpy.management as DM
import ErrorUtils as ERROR
import SSDataObject as SSDO
import SSUtilities as UTILS
import Stats as STATS
import MoransI_Step as GI
import WeightsUtilities as WU
import locale as LOCALE
LOCALE.setlocale(LOCALE.LC_ALL, '')

############## Helper Functions ##############

masterJustify = ["right"] * 6 + ["left"]

dispatchReportType = {"MAX_NUMBER_OF_EXPLANATORY_VARIABLES_ONLY" : 0,
                      "MIN_MAX_NUMBER_OF_EXPLANATORY_VARIABLES_ONLY" : 1,
                      "ALL" : 2, "NONE" : 3}

def returnPerc(numer, denom):
    if numer == 0:
        return 0.0
    else:
        return ( (numer * 1.0) / denom) * 100.

def Combinations(items, n):
    if n==0:
        yield []
    else:
        for i in xrange(len(items)):
            for cc in Combinations(items[i+1:],n-1):
                yield [items[i]]+cc

def runMoransI(ssdo, residuals, swmFile):
    mi = GI.GlobalI_Step(ssdo, residuals, swmFile)
    return mi

def nChooseK(n, k):
    top = MATH.factorial(n)
    left = MATH.factorial(k)
    right = MATH.factorial(n - k)
    return (top * 1.0) / (left * right)

def inSameCombo(n, k):
    top = MATH.factorial(n - 2)
    left = MATH.factorial(k - 2)
    right = MATH.factorial(n - k)
    return (top * 1.0) / (left * right)

################ Interfaces ##################

def runExploratoryRegression():
    """Retrieves the parameters from the User Interface and executes the
    appropriate commands."""

    #### Get User Provided Inputs ####
    ARCPY.env.overwriteOutput = True
    inputFC = ARCPY.GetParameterAsText(0)      
    dependentVar = ARCPY.GetParameterAsText(1).upper()  
    independentVarsReg = ARCPY.GetParameterAsText(2)
    independentVars = independentVarsReg.upper().split(";")
    swmFile = ARCPY.GetParameterAsText(3)
    outputReportFile = ARCPY.GetParameterAsText(4)
    outputTabDir = ARCPY.GetParameterAsText(5)
    outputType = ARCPY.GetParameterAsText(6).upper()
    if outputTabDir == "#" or outputTabDir == "": 
        outputTabDir = None
        outputType = 3
    else:
        outputType = dispatchReportType[outputType]
    maxIndVars = int(ARCPY.GetParameterAsText(7))
    minIndVars = int(ARCPY.GetParameterAsText(8))
    minR2 = LOCALE.atof(ARCPY.GetParameterAsText(9))
    maxCoef = LOCALE.atof(ARCPY.GetParameterAsText(10))
    maxVIF = LOCALE.atof(ARCPY.GetParameterAsText(11))
    minJB = LOCALE.atof(ARCPY.GetParameterAsText(12))
    minMI = LOCALE.atof(ARCPY.GetParameterAsText(13))

    exploreRegress = ExploratoryRegression(inputFC, dependentVar, 
                                           independentVars, swmFile, 
                                           outputReportFile, 
                                           outputTabDir = outputTabDir, 
                                           outputType = outputType, 
                                           maxIndVars = maxIndVars, 
                                           minIndVars = minIndVars, 
                                           minR2 = minR2, maxCoef = maxCoef,
                                           maxVIF = maxVIF, minJB = minJB, 
                                           minMI = minMI)

################## Classes ###################

class ResultHandler(object):
    """Handles result information for Exploratory Regression."""

    def __init__(self, allVarNames, numChoose, ssdo, swmFile,
                 minR2 = .5, maxCoef = .01, maxVIF = 5.0, 
                 minJB = .1, minMI = .1):
        
        #### Set Initial Attributes ####
        UTILS.assignClassAttr(self, locals())

        #### Set Label ####
        self.numVars = len(self.allVarNames)
        self.label = "Choose %i of %i Summary" % (numChoose, self.numVars)
        if numChoose <= 2:
            self.eachAppears = 1
        else:
            self.eachAppears = nChooseK(self.numVars - 2, numChoose - 2)

        #### Set Result Structures ####
        self.signDict = {}
        self.vifDict = {}
        for varName in self.allVarNames:
            self.signDict[varName] = [0, 0]
            self.vifDict[varName] = [0, []]

        self.olsResults = {}
        self.bestR2Vals = []
        self.bestR2Res = []
        self.passTable = []
        self.passBools = []
        self.r2Residuals = NUM.empty((self.ssdo.numObs, 3), dtype = float)

    def runR2Moran(self):
        """Runs Moran's I for highest R2 Models."""
        for ind, olsID in enumerate(self.bestR2Res):
            olsRes = self.olsResults[olsID]
            if olsRes.miPVal == None:
                residuals = self.r2Residuals[:,ind].flatten()
                mi = runMoransI(self.ssdo, residuals, self.swmFile)
                olsRes.setMoransI(mi.pVal)

    def evaluateResult(self, olsResult, residuals, keep = False):
        """Evaluates an OLS result in the context of search criteria."""

        #### Evaluate R2 ####
        r2Value = olsResult.r2
        lenR2 = len(self.bestR2Vals)
        inR2 = False
        if lenR2 < 3:
            self.bestR2Vals.append(r2Value)
            self.bestR2Res.append(olsResult.id)
            self.r2Residuals[:,lenR2] = residuals
            inR2 = True
        else:
            minIndex = NUM.argsort(self.bestR2Vals)[0]
            minValue = self.bestR2Vals[minIndex]
            if r2Value > minValue:
                self.bestR2Vals[minIndex] = r2Value
                self.bestR2Res[minIndex] = olsResult.id
                self.r2Residuals[:,minIndex] = residuals
                inR2 = True

        #### Add to Master List of OLS Results ####
        keepBool = (keep or inR2)
        if keepBool:
            self.olsResults[olsResult.id] = olsResult
        
        #### Evaluate p-values ####
        pValVars = olsResult.evaluatePVals(maxCoef = self.maxCoef)

        #### Evaluate VIF ####
        vifVars = olsResult.evaluateVIF(maxVIF = self.maxVIF)

        #### Populate Result Structures ####
        for varName in olsResult.varNames:
            self.signDict[varName][0] += 1
        for varName in pValVars:
            self.signDict[varName][1] += 1
        for varName in vifVars:
            self.vifDict[varName][0] += 1
            self.vifDict[varName][1] += list(vifVars)

        #### Obtain Bools ####
        pvBool = len(pValVars) == self.numChoose
        vifBool = len(vifVars) == 0
        r2Bool = olsResult.r2 >= self.minR2
        jbBool = olsResult.jb > self.minJB

        #### Decision Based on Bools ####
        tableBool = pvBool and vifBool
        if tableBool:
            self.passTable.append(olsResult.id)

        allBool = pvBool and vifBool and r2Bool and jbBool
        miBool = False
        if allBool:
            mi = runMoransI(self.ssdo, residuals, self.swmFile)
            olsResult.setMoransI(mi.pVal)
            if mi.pVal > self.minMI:
                self.passBools.append(olsResult.id)
                self.olsResults[olsResult.id] = olsResult
                miBool = True

        return r2Bool, pvBool, vifBool, jbBool, miBool, keepBool

    def report(self):
        """Reports the results from exploratory regression analysis."""

        #### Set Title ####
        title = self.label

        #### Column Labels ####
        r2Info = [ ["R2", "AICc", "JB", "BP", "VIF", "MI", "  Model"] ]

        #### Adjusted R2 ####
        header = "Highest Adjusted R-Squared Results"
        sortIndex = NUM.argsort(self.bestR2Vals).tolist()
        sortIndex.reverse()
        for ind in sortIndex:
            olsID = self.bestR2Res[ind]
            olsRes = self.olsResults[olsID]
            olsOut = olsRes.report(formatStr = "%0.2f")
            r2Info.append(olsOut)

        r2Report = UTILS.outputTextTable(r2Info, header = header, 
                                         justify = masterJustify)

        #### Passing Models ####
        header = "Passing Models"
        passList = [ ["R2", "AICc", "JB", "BP", "VIF", "MI", "  Model"] ]
        r2Values = []
        olsIDs = []
        for olsID in self.passBools:
            olsRes = self.olsResults[olsID]
            r2Values.append(olsRes.r2)
            olsIDs.append(olsID)
        sortIndex = NUM.argsort(r2Values).tolist()
        sortIndex.reverse()
        for ind in sortIndex:
            olsID = olsIDs[ind]
            olsRes = self.olsResults[olsID]
            olsOut = olsRes.report(formatStr = "%0.6f")
            passList.append(olsOut)

        passingReport = UTILS.outputTextTable(passList, header = header)

        #### Print Report ####
        starMess = "*" * 78
        finalReport = [starMess, title, r2Report, passingReport]
        finalReport = "\n".join(finalReport)
        finalReport = finalReport + "\n"
        ARCPY.AddMessage(finalReport)

        return finalReport

class OLSResult(object):
    """Holds OLS Result Info for Exploratory Regression."""
    def __init__(self, id, varNames, coef, pVals, vifVals,
                 r2, aic, jb, bp):

        #### Set Initial Attributes ####
        UTILS.assignClassAttr(self, locals())
        self.pVals = NUM.array(pVals)
        self.varNameArray = NUM.array(self.varNames)
        self.miPVal = None
        self.k = len(varNames)

        #### Create Model to Print ####
        self.createModel()

    def evaluateVIF(self, maxVIF = 5.0):
        """Evaluates VIF values."""

        if self.k >= 2:
            self.maxVIFValue = self.vifVals.max()
            overIndices = NUM.where(self.vifVals >= maxVIF)
            return self.varNameArray[overIndices]
        else:
            self.maxVIFValue = 1.0
            return NUM.array([])

    def evaluatePVals(self, maxCoef = .01):
        """Evaluates coefficient p-values."""

        overIndices = NUM.where(self.pVals <= maxCoef)
        return self.varNameArray[overIndices]

    def createModel(self):
        model = []
        for ind, varName in enumerate(self.varNames):
            pVal = self.pVals[ind]
            coefVal = self.coef[ind]

            #### Determine Variable Sign ####
            if coefVal < 0:
                vRes = " -"
            else:
                vRes = " +"

            #### Determine Significance Level ####
            if pVal <= .1 and pVal > .05:
                vRes += varName + "*"
            elif pVal <= .05 and pVal > .01:
                vRes += varName + "**"
            elif pVal <= .01:
                vRes += varName + "***"
            else:
                vRes += varName 

            #### Add to Model ####
            model.append(vRes)

        #### Set Attribute ####
        self.model = " ".join(model)

    def setMoransI(self, value):
        self.miPVal = value

    def report(self, orderType = 0, formatStr = "%0.6f", addModel = True):
        """Reports the results of the OLS run.

        INPUTS:
        orderType {int, 0}: Sort by - 0:R2, 1:Jarque-Bera, 2:Moran
        formatStr (str): format string, E.g. "%0.6f"
        addModel {bool, True}: Add model to report?
        """

        #### Set Output Moran's I p-value ####
        if self.miPVal == None:
            miVal = ""
        else:
            miVal = self.miPVal
        
        if orderType == 0:
            resultList = [ self.r2, self.aic, self.jb, self.bp, 
                           self.maxVIFValue, miVal ]
        elif orderType == 1:
            resultList = [ self.jb, self.r2, self.aic, self.bp, 
                           self.maxVIFValue, miVal ]
        else:
            resultList = [ miVal, self.r2, self.aic, self.jb, self.bp, 
                           self.maxVIFValue ]

        resultList = [UTILS.formatValue(val, formatStr) for val in resultList]
        if addModel:
            resultList.append(self.model)

        return resultList 


class ExploratoryRegression(object):
    """Computes linear regression via Ordinary Least Squares, 
    Psuedo-Step-Wise
    """

    def __init__(self, inputFC, dependentVar, independentVars, swmFile, 
                 outputReportFile, outputTabDir = None, outputType = 0, 
                 maxIndVars = 5, minIndVars = 1, minR2 = .5, 
                 maxCoef = .01, maxVIF = 5.0, minJB = .1, minMI = .1):

        ARCPY.env.overwriteOutput = True

        #### Set Initial Attributes ####
        UTILS.assignClassAttr(self, locals())

        #### Create a Spatial Stats Data Object (SSDO) ####
        self.ssdo = SSDO.SSDataObject(inputFC)

        #### Set Unique ID Field ####
        self.masterField = UTILS.setUniqueIDField(self.ssdo,  
                                      weightsFile = swmFile)

        #### Initialize Data ####
        self.runModels()

    def runModels(self):
        """Performs additional validation and populates the 
        SSDataObject."""

        #### Shorthand Attributes ####
        ssdo = self.ssdo

        #### MasterField Can Not Be The Dependent Variable ####
        if self.masterField == self.dependentVar:
            ARCPY.AddIDMessage("ERROR", 945, self.masterField, 
                               ARCPY.GetIDMessage(84112))
            raise ERROR.ScriptError()

        #### Remove the MasterField from Independent Vars #### 
        if self.masterField in self.independentVars:
            self.independentVars.remove(self.masterField)
            ARCPY.AddIDMessage("Warning", 736, self.masterField)

        #### Remove the Dependent Variable from Independent Vars ####
        if self.dependentVar in self.independentVars:
            self.independentVars.remove(self.dependentVar)
            ARCPY.AddIDMessage("Warning", 850, self.dependentVar)

        #### Raise Error If No Independent Vars ####
        if not len(self.independentVars):
            ARCPY.AddIDMessage("Error", 737)
            raise ERROR.ScriptError()

        #### Obtain Data ####
        self.allVars = [self.dependentVar] + self.independentVars
        ssdo.obtainDataGA(self.masterField, self.allVars, minNumObs = 5)

        #### Create Dependent Variable ####
        self.y = ssdo.fields[self.dependentVar].data
        self.n = ssdo.numObs
        self.y.shape = (self.n, 1)

        #### Assure that Variance is Larger than Zero ####
        yVar = NUM.var(self.y)
        if NUM.isnan(yVar) or yVar <= 0.0:
            ARCPY.AddIDMessage("Error", 906)
            raise ERROR.ScriptError()

        #### Validate Chosen Number of Combos ####
        k = len(ssdo.fields)
        if self.maxIndVars > (k - 1):
            wrn = "Cannot choose combination of size: " 
            wrn += str(self.maxIndVars)
            ARCPY.AddWarning(wrn)
            ARCPY.AddWarning("Setting number of variables to: " + str(k - 1))
            self.maxIndVars = k - 1
        if self.maxIndVars < 1:
            ARCPY.AddWarning("You must choose a combination of at least 1")

        #### Assure Min Vars is less than or equal to Max Vars ####
        if self.maxIndVars <= self.minIndVars:
            wrn = "Minimum number of of independent variables must "
            wrn += "be less than or equal to the maximum...\n"
            wrn += "Setting the minimum to the maximum..."
            self.minIndVars = self.maxIndVars

        #### Gen Range Combos ####
        rangeVars = range(1, k)
        rangeCombos = range(self.minIndVars, self.maxIndVars+1)

        #### Set Output Table Types ####
        if self.outputType == 0:
            self.tables2Create = [self.maxIndVars]
        elif self.outputType == 1:
            self.tables2Create = [self.minIndVars, self.maxIndVars]
        elif self.outputType == 2:
            self.tables2Create = rangeCombos
        else:
            self.tables2Create = []

        #### Create Base Design Matrix ####
        self.x = NUM.ones((self.n, k))
        for column, variable in enumerate(self.independentVars):
            self.x[:,column + 1] = ssdo.fields[variable].data

        #### Calculate Global VIF ####
        if k > 2:
            xTemp = self.x.T[1:]
            corX = NUM.corrcoef(xTemp)
            ic = LA.inv(corX)
            self.globalVifVals = ic.diagonal()
            self.printVIF = True
        else:
            self.globalVifVals = None
            self.printVIF = False

        #### Create Base Table Field Names ####
        if len(self.tables2Create):
            self.outFieldNames = ["RUNID", "AdjR2", "AICc", "JB", 
                                  "BP", "MaxVIF", "SA"]
            self.outFieldTypes = ["LONG", "DOUBLE", "DOUBLE", "DOUBLE", 
                                  "DOUBLE", "DOUBLE", "DOUBLE"]
            fieldLens = [ len(i) for i in self.independentVars ]
            self.maxFieldLen = max(fieldLens) + 5

            #### Create Table Prefix Based on Output Text File Prefix ####
            self.tablePrefix = UTILS.getBasenamePrefix(self.outputReportFile)

            #### Set NULL Values ####
            tableDirType = UTILS.getBaseWorkspaceType(self.outputTabDir)
            if tableDirType.upper() == "FILESYSTEM":
                self.nullValue = UTILS.shpFileNull
            else:
                self.nullValue = NUM.nan

        #### Create Output Report File ####
        fo = open(self.outputReportFile, "w")

        #### Hold Results for Every Choose Combo ####
        self.resultDict = {}
        self.sumRuns = 0
        self.sumGI = 0
        self.boolGI = 0
        self.boolResults = NUM.zeros(4, dtype = int)
        self.jbModels = []
        self.jbValues = []
        self.jbResiduals = NUM.empty((self.n, 3), dtype = float)

        for choose in rangeCombos:
            #### Generate Index Combos ####
            comboGenerator = Combinations(rangeVars, choose)

            #### Set Progressor ####
            message = str(k-1) + " variables, choose: "+ str(choose)
            ARCPY.SetProgressor("default", message)
            
            #### Set Result Structure ####
            rh = ResultHandler(self.independentVars, choose, 
                               self.ssdo, self.swmFile,
                               minR2 = self.minR2, maxCoef = self.maxCoef, 
                               maxVIF = self.maxVIF, minJB = self.minJB, 
                               minMI = self.minMI)

            #### Set Bool for Creating Output Table ####
            createTableBool = choose in self.tables2Create

            #### Set Output Table Only Information ####
            if createTableBool:
                tableReportCount = 1
                tableResults = []
                tableNameList = [ "X" + str(j) for j in range(1, choose+1) ]
                self.fieldNamesForTable = self.outFieldNames + tableNameList
                xTypeList = [ "TEXT" for name in tableNameList ]
                self.fieldTypesForTable = self.outFieldTypes + xTypeList

            #### Loop Through All Combinations ####
            modelCount = 0
            for combo in comboGenerator:
                #### Create Design Matrix for Given Combination ####
                columns = [0] + combo
                comboX = self.x[0:,columns]

                #### Get Model Info for given Combination ####
                N, K = NUM.shape(comboX)
                varNameList = [ self.independentVars[j-1] for j in combo ] 
                varNameListInt = ["Intercept"] + varNameList 
                modelAll = str(self.dependentVar) + " ~ "
                modelAll += " + ".join(varNameListInt)
                modelID = str(K) + ":" + str(modelCount)

                #### Run Linear Regression ####
                runModel = self.calculate(comboX)

                if not runModel:
                    msg = "Perfect Multicollinearity in: " + modelAll
                    ARCPY.AddWarning(msg)
                else:
                    #### Keep Track of Total Number of Models Ran ####
                    modelCount += 1
                    self.sumRuns += 1
                    residuals = self.residuals.flatten()

                    #### Evaluate p-values ####
                    if self.BPProb < .1:
                        #### Use Robust Coefficients ####
                        pValsOut = self.pValsRob[1:]
                    else:
                        pValsOut = self.pVals[1:]
                    coefOut = self.coef[1:]

                    #### Set OLS Result ####
                    res = OLSResult(modelID, varNameList, coefOut, pValsOut, 
                                    self.vifVal, self.r2Adj, self.aicc,
                                    self.JBProb, self.BPProb)

                    #### Evaluate Jarque-Bera Stat ####
                    keep = self.pushPopJB(res, self.residuals.flatten())

                    boolReport = rh.evaluateResult(res, residuals, keep = keep) 
                    r2Bool, pvBool, vifBool, jbBool, giBool, keepBool = boolReport

                    #### Populate Output Table ####
                    if createTableBool:
                        lesserKeepModel = pvBool and vifBool
                        if lesserKeepModel:
                            #### Set VIF/Moran's I for Table ####
                            maxVIFValue = res.maxVIFValue
                            giPVal = res.miPVal
                            if giPVal == None:
                                giPVal = self.nullValue

                            #### Create List of Results ####
                            resultValues = [tableReportCount, self.r2Adj, 
                                            self.aicc, self.JBProb, 
                                            self.BPProb, maxVIFValue,
                                            giPVal]
                            resultValues += varNameList
                            tableResults.append(resultValues) 
                            tableReportCount += 1

                    #### Add Booleans for End Total Summary ####
                    boolResult = [r2Bool, pvBool, vifBool, jbBool]
                    self.boolResults += boolResult

                    #### Delete OLS Instance if Not Necessary For Summary ####
                    if not keepBool:
                        del res


            #### Run Moran's I for Highest Adj. R2 ####
            rh.runR2Moran()

            #### Add Results to Report File ####
            result2Print = rh.report()
            fo.write(result2Print)

            #### Add Choose Run to Result Dictionary ####
            self.resultDict[choose] = rh

            #### Create Output Table ####
            if createTableBool:
                self.createOutputTable(choose, tableResults)

        #### Run Moran's I on Best Jarque-Bera ####
        self.createJBReport()

        #### Final Moran Stats ####
        self.getMoranStats()

        #### Ending Summary ####
        self.endSummary()
        fo.write(self.fullReport)

        #### Clean Up ####
        fo.close()

    def createOutputTable(self, choose, tableResults):
        """Create output table for a given set of results."""

        #### Message and Progressor ####
        msg = ARCPY.GetIDMessage(84008)
        ARCPY.AddMessage(msg)
        ARCPY.SetProgressor("default", msg)

        #### Create/Finalize Output Table Name ####
        outName = self.tablePrefix + str(choose)
        outputTable = OS.path.join(self.outputTabDir, outName)
        outputTable, dbf = UTILS.returnTableName(outputTable)
        UTILS.passiveDelete(outputTable)
        outPath, outName = OS.path.split(outputTable)
        DM.CreateTable(outPath, outName)

        #### Add Output Fields ####
        for ind, fieldName in enumerate(self.fieldNamesForTable):
            UTILS.addEmptyField(outputTable, fieldName,
                                self.fieldTypesForTable[ind],
                                length = self.maxFieldLen)

        #### Create Insert Cursor ####
        insert = ARCPY.InsertCursor(outputTable)

        #### Populate Output Table ####
        for result in tableResults:

            #### Create New Row in Table ####
            row = insert.newRow()

            #### Populate the Fields in the Table ####
            eLoop = enumerate(self.fieldNamesForTable)
            for ind, fieldName in eLoop:
                row.setValue(fieldName, result[ind])
            insert.insertRow(row)

        #### Clean Up ####
        ARCPY.AddMessage(outputTable)
        del insert

    def getMoranStats(self):
        self.sumMoranRuns = 0
        self.sumMoranPass = 0
        miValues = []
        miModels = []
        for resKey, resHandler in self.resultDict.iteritems():
            for olsKey, olsRes in resHandler.olsResults.iteritems():
                miValue = olsRes.miPVal
                if miValue != None:
                    if len(miValues) < 3:
                        miValues.append(miValue)
                        miModels.append(olsRes)
                    else:
                        minIndex = NUM.argsort(miValues)[0]
                        minValue = miValues[minIndex]
                        if minValue < miValue:
                            miValues[minIndex] = miValue
                            miModels[minIndex] = olsRes
                    self.sumMoranRuns += 1
                    if miValue > self.minMI:
                        self.sumMoranPass += 1

        self.miReportRows = [ i.report(orderType = 2) for i in miModels ]

    def createJBReport(self):
        self.jbReportRows = []
        sortIndex = NUM.argsort(self.jbValues).tolist()
        sortIndex.reverse()
        for ind in sortIndex:
            olsRes = self.jbModels[ind]
            if olsRes.miPVal == None:
                residuals = self.jbResiduals[:,ind].flatten()
                mi = runMoransI(self.ssdo, residuals, self.swmFile)
                olsRes.setMoransI(mi.pVal)
            
            olsOut = olsRes.report(orderType = 1)
            self.jbReportRows.append(olsOut)

    def pushPopJB(self, olsRes, residuals):
        """Keeps track of the best (highest) Jarque-Bera p-values."""
        lenRes = len(self.jbValues)
        keep = False
        if lenRes < 3:
            self.jbValues.append(self.JBProb)
            self.jbModels.append(olsRes)
            self.jbResiduals[:,lenRes] = residuals
            keep = True
        else:
            minIndex = NUM.argsort(self.jbValues)[0]
            minValue = self.jbValues[minIndex]
            if minValue < self.JBProb:
                self.jbValues[minIndex] = self.JBProb
                self.jbModels[minIndex] = olsRes
                self.jbResiduals[:,minIndex] = residuals
                keep = True
        return keep

    def endSummary(self):
        """Creates End Summary for Report File."""

        #### Passing Model Global Summary ####
        passHeader = "Percentage of Search Criteria Passed"
        passResults = [ ["Search Criterion", "Cutoff", "Trials", 
                         "# Passed", "% Passed"] ]

        cutoffList = [ "> " + UTILS.formatValue(self.minR2, "%0.2f"),
                       "< " + UTILS.formatValue(self.maxCoef, "%0.2f"),
                       "< " + UTILS.formatValue(self.maxVIF, "%0.2f"),
                       "> " + UTILS.formatValue(self.minJB, "%0.2f"),
                       "> " + UTILS.formatValue(self.minMI, "%0.2f") ]

        categories = ["Min Adjusted R-Squared", "Max Coefficient p-value", 
                      "Max VIF Value", "Min Jarque-Bera p-value",
                      "Min Moran's I p-value" ]

        boolPerc = [ returnPerc(i, self.sumRuns) for i in self.boolResults ]
        boolPerc.append( returnPerc(self.sumMoranPass, self.sumMoranRuns) )
        boolOut = list(self.boolResults) + [self.sumMoranPass]
        sumOut = [ self.sumRuns for i in self.boolResults ]
        sumOut += [self.sumMoranRuns]

        for ind, category in enumerate(categories):
            outValue = LOCALE.format("%0.2f", boolPerc[ind])
            outCutoff = cutoffList[ind]
            outTrial = sumOut[ind]
            outCount = boolOut[ind]
            passResults.append( [category, outCutoff, outTrial, 
                                 outCount, outValue] )

        self.passReport = UTILS.outputTextTable(passResults, 
                                        header = passHeader,
                                        pad = 1, justify = "right")

        #### Variable Significance and VIF Reports ####
        #### Create Table Headers ####
        signHeader = "Summary of Variable Significance"
        vifHeader = "Summary of Multicollinearity"

        #### Create Column Labels and Result Lists ####
        signResults = [ ["Variable", "% Significant"] ]
        vifResults = [ ["Variable", "VIF", "Violations",
                        "Covariates"] ]

        #### Get Covariate Total ####
        totalTogether = 0
        for resultKey, result in self.resultDict.iteritems():
            totalTogether += result.eachAppears

        #### Populate Result Lists ####
        for ind, varName in enumerate(self.independentVars):
            totalSign = 0
            totalRan = 0
            totalViolations = 0
            totalCovariates = {}
            for resultKey, result in self.resultDict.iteritems():
                #### Significance Results ####
                numRan, numSign = result.signDict[varName]
                totalSign += numSign 
                totalRan += numRan

                #### VIF Results ####
                numViolate, covariates = result.vifDict[varName]
                totalViolations += numViolate
                for covariate in covariates:
                    if covariate != varName:
                        try:
                            totalCovariates[covariate] += 1
                        except:
                            totalCovariates[covariate] = 1

            #### Complete Significance Row ####
            percentSign = ((totalSign * 1.0) / totalRan) * 100.0
            rowRes = [varName, LOCALE.format("%0.2f", percentSign)]
            signResults.append(rowRes)

            #### Complete VIF Row ####
            globalVIF = self.globalVifVals[ind]
            if abs(globalVIF) > 1000:
                globalVIFOut = "> " + LOCALE.format("%0.2f", 1000.)
            else:
                globalVIFOut = LOCALE.format("%0.2f", globalVIF)
            coString = []
            for covariate, totalTimes in totalCovariates.iteritems():
                tRatio = (totalTimes/totalTogether) * 100
                tRatio = LOCALE.format("%0.2f", tRatio)
                coString.append(covariate + " (" + str(tRatio) + ")")
            if len(coString):
                coString = ", ".join(coString)
            else:
                coString = "--------"
            vifRes = [varName, globalVIFOut, "%i" % totalViolations, coString]
            vifResults.append(vifRes)

        #### Create Report Tables ####
        self.signReport = UTILS.outputTextTable(signResults, 
                                                header = signHeader,
                                                justify = ["left", "right"],
                                                pad = 1)

        self.vifReport = UTILS.outputTextTable(vifResults, header = vifHeader,
                                               justify = ["left", "right",
                                                          "center", "left"],
                                               pad = 1)

        #### Residual Normality Summary ####
        jbHeader = "Summary of Residual Normality"
        jbResults = [ ["JB", "R2", "AICc", "BP", "VIF", "MI", "  Model"] ]
        jbResults += self.jbReportRows
        self.jbReport = UTILS.outputTextTable(jbResults, header = jbHeader,
                                          pad = 1, justify = masterJustify)

        #### Residual Autocorrelation ####
        miHeader = "Summary of Residual Autocorrelation"
        miResults = [ ["MI", "R2", "AICc", "JB", "BP", "VIF", "  Model"] ]
        miResults += self.miReportRows
        justify = ["right"] * 6 + ["left"]
        self.miReport = UTILS.outputTextTable(miResults, header = miHeader,
                                          pad = 1, justify = masterJustify)

        #### Abbreviation Table ####
        abbHeader = "Table Abbreviations"
        modelString1 = "Variable sign and significance"
        modelString2 = "* = 0.10, ** = 0.05, *** = 0.01"
        abbResults = [ ["R2", "Adjusted R-Squared"],
                       ["AICc", "Akaike's Information Criterion"],
                       ["JB", "Jarque-Bera p-value"],
                       ["BP", "Koenker (BP) Statistic p-value"],
                       ["VIF", "Max Variance Inflation Factor"],
                       ["MI", "Moran's I p-value"],
                       ["Model", modelString1],
                       ["", modelString2]]

        self.abbReport = UTILS.outputTextTable(abbResults, header = abbHeader,
                                               pad = 1)

        #### Display Tables ####
        starMess = "*" * 78
        dashMess = "-" * 78
        ARCPY.AddMessage(starMess)
        globalHeader = " Exploratory Regression Global Summary (%s) "
        globalHeader = globalHeader % self.dependentVar
        globalHeader = globalHeader.center(78, "*")
        ARCPY.AddMessage(globalHeader)
        ARCPY.AddMessage(self.passReport)
        ARCPY.AddMessage(dashMess)
        ARCPY.AddMessage(self.signReport)
        ARCPY.AddMessage(dashMess)
        ARCPY.AddMessage(self.vifReport)
        ARCPY.AddMessage(dashMess)
        ARCPY.AddMessage(self.jbReport)
        ARCPY.AddMessage(dashMess)
        ARCPY.AddMessage(self.miReport)
        ARCPY.AddMessage(dashMess)
        ARCPY.AddMessage(self.abbReport)
        ARCPY.AddMessage(dashMess)
        self.fullReport = [starMess, globalHeader, self.passReport,
                           dashMess, self.signReport,
                           dashMess, self.vifReport, 
                           dashMess, self.jbReport,
                           dashMess, self.miReport,
                           dashMess, self.abbReport,
                           dashMess]
        self.fullReport = "\n".join(self.fullReport)

    def calculate(self, comboX):
        """Performs OLS and related diagnostics."""

        #### Shorthand Attributes ####
        ssdo = self.ssdo
        x = comboX
        n, k = NUM.shape(comboX)
        y = self.y

        #### General Information ####
        fn = n * 1.0
        dof = n - k
        fdof = dof * 1.0
        xt = x.T        
        yt = y.T
        xx = NUM.dot(xt, x)

        #### Attempt to Invert Design Matrix ####
        try: 
            xxi = LA.inv(xx)
        except:
            #### Perfect multicollinearity, cannot proceed ####
            ARCPY.AddIDMessage("Warning", 639)
            return False

        #### Bad Probabilities - Near Multicollinearity ####
        badProbs = False
        
        #### Compute Coefficients ####
        xy = NUM.dot(xt, y) 
        coef = NUM.dot(xxi, xy)

        #### Residuals, Sum Of Squares, R2, Etc. ####
        yHat = NUM.dot(x, coef)
        yBar = (y.sum())/fn
        e = y - yHat
        ess = ( NUM.dot(e.T, e) )[0][0]
        s2 = (ess / fdof)
        s2mle = (ess / fn)
        seResiduals = NUM.sqrt(s2)
        ss = y - yBar
        tss = ( NUM.dot(ss.T, ss) )[0][0]
        r2 = 1.0 - (ess/tss) 
        r2Adj =  1.0 - ( (ess / (fdof)) / (tss / (fn-1)) )  
        u2 = e * e

        #### Variance-Covariance for Coefficients ####
        varBeta = xxi * s2

        #### Standard Errors / t-Statistics ####
        seBeta = NUM.sqrt(varBeta.diagonal())
        tStat = (coef.T / seBeta).flatten()

        #### White's Robust Standard Errors ####
        dofScale =  ( n / (n - k) ) * 1.0 
        sHat = NUM.dot((u2 * x).T, x) * dofScale
        varBetaRob = NUM.dot(NUM.dot(xxi, sHat), xxi)   
        seBetaRob =  NUM.sqrt(varBetaRob.diagonal())
        tStatRob = (coef.T / seBetaRob).flatten()

        #### Coefficient t-Tests ####
        pVals = []
        pValsRob = []
        for varInd in xrange(k):
            #### General ####
            try:
                p = STATS.tProb(tStat[varInd], dof, type = 2)
            except:
                p =  NUM.nan
                badProbs = True
            pVals.append(p)

            #### Robust ####
            try:
                p = STATS.tProb(tStatRob[varInd], dof, type = 2)
            except:
                p =  NUM.nan
                badProbs = True
            pValsRob.append(p)

        #### Jarque-Bera Test For Normality of the Residuals ####
        muE = (e.sum()) / fn
        devE = e - muE
        u3 = (devE**3.0).sum() / fn
        u4 = (devE**4.0).sum() / fn
        denomS = s2mle**1.5
        denomK = s2mle**2.0
        skew = u3 / denomS
        kurt = u4 / denomK
        self.JB = (n/6.) * ( skew**2. + ( (kurt - 3.)**2. / 4. ))
        if self.JB >= 0.0:
            self.JBProb = STATS.chiProb(self.JB, 2, type = 1)
        else:
            self.JBProb = NUM.nan
            badProbs = True

        #### Breusch-Pagan Test for Heteroskedasticity ####
        u2y = NUM.dot(xt, u2)
        bpCoef = NUM.dot(xxi, u2y)
        u2Hat = NUM.dot(x, bpCoef)
        eU = u2 - u2Hat
        essU = NUM.dot(eU.T, eU)
        u2Bar = (u2.sum()) / fn
        ssU = u2 - u2Bar
        tssU = NUM.dot(ssU.T, ssU)
        r2U = 1.0 - (essU/tssU)
        self.BP = (fn * r2U)[0][0]
        if self.BP >= 0.0:
            self.BPProb = STATS.chiProb(self.BP, (k-1), type = 1)
        else:
            self.BPProb = NUM.nan
            badProbs = True

        #### Classic Joint-Hypothesis F-Test ####
        q = k - 1
        fq = q * 1.0
        self.fStat = (r2/fq) / ((1 - r2) / (fn - k))
        try:
            self.fProb = abs(STATS.fProb(self.fStat, q, 
                                      (n-k), type = 1))
        except:
            self.fProb = NUM.nan
            badProbs = True

        #### Wald Robust Joint Hypothesis Test ####
        R = NUM.zeros((q,k))
        R[0:,1:] = NUM.eye(q)
        Rb = NUM.dot(R, coef) 

        try:
            invRbR = LA.inv( NUM.dot(NUM.dot(R, varBetaRob), R.T) )
        except:
            #### Perfect multicollinearity, cannot proceed ####
            ARCPY.AddIDMessage("Warning", 639)
            return False

        self.waldStat = ( NUM.dot(NUM.dot(Rb.T, invRbR), Rb) )[0][0]
        if self.waldStat >= 0.0:
            self.waldProb = STATS.chiProb(self.waldStat, q, type = 1)
        else:
            self.waldProb = NUM.nan
            badProbs = True

        #### Log-Likelihood ####
        self.logLik = -(n / 2.) * (1. + NUM.log(2. * NUM.pi)) - \
                       (n / 2.) * NUM.log(s2mle)

        #### AIC/AICc ####
        k1 = k + 1
        self.aic = -2. * self.logLik + 2. * k1
        self.aicc = -2. * self.logLik + 2. * k1 * (fn / (fn - k1 - 1))

        #### Calculate the Variance Inflation Factor ####
        if k <= 2:
            self.vifVal = ARCPY.GetIDMessage(84090)
            self.vif = False
        else:
            xTemp = xt[1:]
            corX = NUM.corrcoef(xTemp)
            try:
                ic = LA.inv(corX)
                self.vifVal = ic.diagonal()
                self.vif = True 
            except:
                #### Perfect multicollinearity, cannot proceed ####
                ARCPY.AddIDMessage("Warning", 639)
                return False

        #### Set Attributes ####
        self.dof = dof
        self.coef = coef
        self.yHat = yHat
        self.yBar = yBar
        self.residuals = e
        self.seResiduals = seResiduals
        self.stdRedisuals = e / self.seResiduals
        self.ess = ess
        self.tss = tss
        self.varCoef = varBeta
        self.seCoef = seBeta
        self.tStats = tStat
        self.pVals = pVals
        self.varCoefRob = varBetaRob
        self.seCoefRob = seBetaRob
        self.tStatsRob = tStatRob
        self.pValsRob = pValsRob
        self.r2 = r2
        self.r2Adj = r2Adj
        self.s2 = s2
        self.s2mle = s2mle
        self.q = q
        self.badProbs = badProbs
        self.varLabels = [ARCPY.GetIDMessage(84064)] + self.independentVars

        return True

if __name__ == '__main__':
    er = runExploratoryRegression()



