"""
Tool Name:     Global Moran's I
Source Name:   GlobalI.py
Version:       ArcGIS 10.0
Author:        Environmental Systems Research Institute Inc.
Description:   Computes Global Moran's I statistic
"""

################### Imports ########################
import sys as SYS
import os as OS
import numpy as NUM
import arcpy as ARCPY
import arcpy.management as DM
import ErrorUtils as ERROR
import SSUtilities as UTILS
import SSDataObject as SSDO
import Stats as STATS
import WeightsUtilities as WU
import gapy as GAPY
import locale as LOCALE
LOCALE.setlocale(LOCALE.LC_ALL, '')

def setupGlobalI_Increment():
    """Retrieves the parameters from the User Interface and executes the
    appropriate commands."""

    #### Input Features and Variable ####
    inputFC = ARCPY.GetParameterAsText(0)                    
    varName = ARCPY.GetParameterAsText(1).upper()
    
    #### Number of Distance Thresholds ####
    nIncrements = ARCPY.GetParameterAsText(2)
    if nIncrements == "#" or nIncrements == "": 
        nIncrements = None
    else: 
        nIncrements = int(nIncrements)
        if nIncrements > 30:
            nIncrements = 30
    
    #### Starting Distance ####
    begDist = ARCPY.GetParameterAsText(3)
    if begDist == "#" or begDist == "": 
        begDist = None
    else: 
        begDist = LOCALE.atof(begDist)

    #### Step Distance ####
    dIncrement = ARCPY.GetParameterAsText(4)
    if dIncrement == "#" or dIncrement == "": 
        dIncrement = None
    else: 
        dIncrement = LOCALE.atof(dIncrement)

    #### EUCLIDEAN or MANHATTAN ####
    distanceConcept = ARCPY.GetParameterAsText(5).upper().replace(" ", "_")
    concept = WU.conceptDispatch[distanceConcept]

    #### Row Standardized ####
    rowStandard = ARCPY.GetParameterAsText(6)
    if rowStandard == 'true':
        rowStandard = True
    else:
        rowStandard = False

    #### Output Table ####
    outputTable = ARCPY.GetParameterAsText(7)
    if outputTable == "#" or outputTable == "": 
        outputTable = None

    #### Display Graph ####
    displayIt = ARCPY.GetParameterAsText(8)
    if displayIt == 'true':
        displayIt = True
    else:
        displayIt = False

    #### Run Analysis ####
    gi = GlobalI_Step(inputFC, varName, nIncrements = nIncrements, 
                      begDist = begDist, dIncrement = dIncrement,
                      concept = concept, rowStandard = rowStandard)

    #### Report Results ####
    reportTable = gi.report()

    #### Optionally Create Output ####
    if outputTable:
        gi.createOutput(outputTable, displayIt = displayIt)

class GlobalI_Step(object):
    """Calculates Global Morans I, Incremental Distance Version:

    INPUTS:
    inputFC (str): path to the input feature class
    varName (str): name of analysis field
    nIncrements {int, 10}: number of distance bands
    begDist {float, None}: starting distance for analysis
    dIncrement {float, None}: increase in distance per increment
    concept: {str, EUCLIDEAN}: EUCLIDEAN or MANHATTAN 
    rowStandard {bool, True}: row standardize weights?

    ATTRIBUTES:
    numObs (int): number of features in analysis
    y (array, numObs x 1): vector of field values

    METHODS:
    report: reports results as a printed message or to a file
    createOutput: creates an output table for Moran's I results
    """

    def __init__(self, inputFC, varName, nIncrements = 10, 
                 begDist = None, dIncrement = None,
                 concept = "EUCLIDEAN", rowStandard = True):

        #### Set Initial Attributes ####
        UTILS.assignClassAttr(self, locals())

        #### Create a Spatial Stats Data Object (SSDO) ####
        self.ssdo = SSDO.SSDataObject(inputFC)

        #### Set Unique ID Field ####
        self.masterField = UTILS.setUniqueIDField(self.ssdo)

        #### Initialize Data ####
        self.initialize()

        #### Construct Based on SWM File or On The Fly ####
        self.construct()

    def initialize(self):
        """Populates the instance of the Spatial Statistics Data 
        Object (SSDataObject) and resolves a default distance threshold
        if none given.
        """

        #### Shorthand Attributes ####
        ssdo = self.ssdo
        varName = self.varName
        self.master2Order = ssdo.master2Order
        masterField = self.masterField
        concept = self.concept
        gaExtent = UTILS.get92Extent(ssdo.extent)

        #### Populate SSDO with Data ####
        ssdo.obtainDataGA(masterField, [varName], minNumObs = 3, 
                          warnNumObs = 30)
        field = ssdo.fields[varName]
        self.y = field.data
        self.numObs = ssdo.numObs

        #### Set Extent and Envelope ####
        self.envelope = UTILS.Envelope(ssdo.extent)
        envCoords = self.envelope.envelope
        self.maxDistance = self.envelope.maxExtent * 0.25
        self.minX, self.minY, self.maxX, self.maxY = envCoords

        #### Determine Starting Distance ####
        if self.begDist == None:
            self.begDist = WU.createThresholdDist(ssdo.gaTable, 
                                                  concept = concept)

        #### Negative Values Not Valid ####
        if self.begDist < 0:
            ARCPY.AddIDMessage("ERROR", 933)
            raise ERROR.ScriptError()

        #### Determine Distance Increment ####
        if not self.dIncrement:
            if self.begDist: 
                distRange = self.maxDistance - self.begDist
            else: 
                distRange = self.maxDistance
            self.dIncrement = float(distRange / self.nIncrements)

        #### Determine All Distance Cutoffs ####
        rangeInc = xrange(self.nIncrements)
        cutoffs = []
        for inc in rangeInc:
            val = (inc * self.dIncrement) + self.begDist

            #### Assure All Features Are Not Related ####
            #### Uses Inverse Distance to Avoid ScriptError() ####
            threshold, maxSet = WU.checkDistanceThreshold(gaExtent, val,
                                                         weightType = 0)
            
            #### Add Cutoff if Appropriate ####
            if not maxSet:
                cutoffs.append(val)

        #### Check Cutoff Values ###
        stepMax = cutoffs[-1]
        if self.begDist > (self.envelope.maxExtent * 0.51):
            ARCPY.AddIDMessage("WARNING", 934)
        elif stepMax > self.envelope.maxExtent:
            ARCPY.AddIDMessage("WARNING", 935)

        #### Create Results Labels and Field Names ####
        resLabels = [ARCPY.GetIDMessage(84179), ARCPY.GetIDMessage(84148), 
                     ARCPY.GetIDMessage(84149), ARCPY.GetIDMessage(84150), 
                     ARCPY.GetIDMessage(84151), ARCPY.GetIDMessage(84152)]
        self.resLabels = [ i.replace(":", "").strip() for i in resLabels ]
        self.outputFields = [self.resLabels[0], "MoransI", "ExpectedI",
                             self.resLabels[3], self.resLabels[4], 
                             self.resLabels[5]]

        #### Set Attributes ####
        self.stepMax = stepMax
        self.cutoffs = NUM.array(cutoffs)
        self.cutoffOrder = range(self.nIncrements)
        self.largestDistBand = self.cutoffs[-1]

    def construct(self):
        """Constructs the neighborhood structure for each feature and
        dispatches the appropriate values for the calculation of the
        statistic."""

        #### Shorthand Attributes ####
        ssdo = self.ssdo
        numObs = ssdo.numObs
        master2Order = self.master2Order
        masterField = self.masterField
        concept = self.concept
        iterVals = range(numObs)
        rowStandard = self.rowStandard
        wType = 1

        yVar = NUM.var(self.y)
        if NUM.isnan(yVar) or yVar <= 0.0:
            ARCPY.AddIDMessage("Error", 906)
            raise ERROR.ScriptError()

        #### Create Deviation Variables ####
        self.yBar = NUM.mean(self.y)
        self.yDev = self.y - self.yBar

        #### Create Results Array ####
        self.giResults = NUM.zeros((self.nIncrements, 6))

        #### Run Analysis at Varying Distances ####
        msgProg = "Running Moran's I at Varying Distances...."
        ARCPY.SetProgressor("step", msgProg, 0, self.nIncrements, 1)
        ARCPY.AddMessage("\n" + msgProg)
        gaTable = ssdo.gaTable

        for ind, cutoff in enumerate(self.cutoffs):
            msg = "%i of %i: %0.4f" 
            ARCPY.AddMessage(msg % ((ind + 1), self.nIncrements, cutoff))
            gaSearch = GAPY.ga_nsearch(gaTable)
            gaSearch.init_nearest(cutoff, 0, concept.lower())

            #### Create Base Data Structures/Variables #### 
            self.numer = 0.0
            self.denom = NUM.sum(self.yDev**2.0)
            self.rowSum = NUM.zeros(numObs)
            self.colSum = NUM.zeros(numObs)
            self.s0 = 0
            self.s1 = 0
            self.wij = {}

            #### Keep Track of Features with No Neighbors ####
            self.idsNoNeighs = []

            #### Calculation For Each Feature ####
            for i in iterVals:
                masterID = gaTable[i][0]
                rowInfo = WU.getWeightsValuesOTF(i, master2Order, self.yDev,
                                                 gaTable, gaSearch, wType,
                                                 concept = concept, 
                                                 threshold = cutoff, 
                                                 rowStandard = rowStandard)

                #### Parse Row Info ####
                orderID, yiDev, nhIDs, nhVals, weights = rowInfo

                #### Assure Neighbors Exist After Selection ####
                nn = len(nhVals)

                if nn:
                    #### Process Feature Contribution to Moran's I ####
                    self.processRow(orderID, yiDev, nhIDs, 
                                          nhVals, weights) 

                else:
                    #### No Neighbors ####
                    self.addNoNeighbor(masterID, orderID)

            #### Report if All Features Have No Neighbors ####
            if self.s0 == 0:
                ARCPY.AddIDMessage("Error", 908)
                raise ERROR.ScriptError()

            #### Report on Features with No Neighbors ####
            noNeighs = len(self.idsNoNeighs)
            if noNeighs:
                self.idsNoNeighs.sort()
                if noNeighs > 30:
                    printNoNeighs = self.idsNoNeighs[0:30]
                else:
                    printNoNeighs = self.idsNoNeighs
                ERROR.warningNoNeighbors(numObs, noNeighs, printNoNeighs,
                                         masterField)

            #### Calculate Moran's I ####
            self.calculate()

            #### Pack Results ####
            res = (cutoff, self.gi, self.ei, self.vi, self.zi, self.pVal) 
            self.giResults[ind] = res

            ARCPY.SetProgressorPosition()
        
        ARCPY.AddMessage("\n")

    def processRow(self, orderID, yiDev, nhIDs, nhVals, weights):
        """Processes a features contribution to the Moran's I statistic.
        
        INPUTS:
        orderID (int): order in corresponding numpy value arrays
        yiVal (float): value for given feature
        nhIDs (array, nn): neighbor order in corresponding numpy value arrays
        nhVals (array, nn): values for neighboring features (1)
        weights (array, nn): weight values for neighboring features (1)

        NOTES:
        (1)  nn is equal to the number of neighboring features
        """

        #### Numerator Calculation ####
        sumW = weights.sum()
        self.s0 += sumW
        self.numer += NUM.sum(nhVals * weights) * yiDev

        #### Weights Charactersitics Update ####
        c = 0
        for neighID in nhIDs:
            ij = (orderID, neighID)
            ji = (neighID, orderID)
            w = weights[c] 
            self.s1 += w**2.0
            try:
                self.s1 += 2.0 * w * self.wij.pop(ji)
            except:
                self.wij[ij] = w
            self.rowSum[orderID] += w
            self.colSum[neighID] += w
            c += 1

    def calculate(self):
        """Calculate Moran's I Statistic."""

        s0 = self.s0
        s1 = self.s1
        n = len(self.rowSum) * 1.0
        s2 = NUM.sum( (self.rowSum + self.colSum)**2.0 )
        self.s2 = s2
        self.ei = -1. / (n - 1)
        self.squareExpectedI = self.ei**2
        self.n = n
        scale = n / s0
        s02 = s0 * s0
        n2 = n * n
        n2s1 = n2 * s1
        ns2 = n * s2
        self.gi = (scale * (self.numer/self.denom))
        yDev4Sum = NUM.sum(self.yDev**4) / n
        yDevsqsq = (self.denom / n)**2
        b2 = yDev4Sum / yDevsqsq
        self.b2 = b2
        left = n * ((n2 - (3*n) + 3) * s1 - (n*s2) + 3 * (s02))
        right = b2 * ((n2 - n) * s1 - (2*n*s2) + 6 * (s02))
        denom = (n-1) * (n-2) * (n-3) * s02
        num = (left - right) / denom
        self.expectedSquaredI = num 
        self.vi = self.expectedSquaredI - self.squareExpectedI

        #### Assure that Variance is Larger than Zero ####
        if NUM.isnan(self.vi) or self.vi <= 0.0:
            ARCPY.AddIDMessage("Error", 906)
            raise ERROR.ScriptError()

        self.standDev = NUM.sqrt(self.vi)
        self.zi = (self.gi - self.ei)/self.standDev
        self.pVal = STATS.zProb(self.zi, type = 2)

    def addNoNeighbor(self, masterID, orderID):
        """Accounting for features with no neighbors.
        
        INPUTS:
        masterID (int): unique ID value that has no neighbors
        orderID (int): order in corresponding numpy value arrays
        """

        self.idsNoNeighs.append(masterID)

    def report(self, fileName = None):
        """Reports the Moran's I results as a message or to a file.  If
        self.displayIt is set to True, then an html graphical report is
        generated to your default temp directory.

        INPUTS:
        fileName {str, None}: path to a text file to populate with results.
        """

        header = ARCPY.GetIDMessage(84160) + " by Distance"
        results = [ self.resLabels ] 

        for testIter in range(self.nIncrements):
            d, gi, ei, vi, zi, pv = self.giResults[testIter]
            d = LOCALE.format("%0.2f", round(d, 2))
            gi = LOCALE.format("%0.6f", gi)
            ei = LOCALE.format("%0.6f", ei)
            vi = LOCALE.format("%0.6f", vi)
            zi = LOCALE.format("%0.6f", zi)
            pv = LOCALE.format("%0.6f", pv)
            results.append( [d, gi, ei, vi, zi, pv] ) 

        outputReport = UTILS.outputTextTable(results, header = header, 
                                             justify = "center", pad = 1)
        if fileName:
            f = open(fileName, "w")
            f.write(outputReport)
            f.close()
        else:
            ARCPY.AddMessage(outputReport)

        return outputReport

    def createOutput(self, outputTable, displayIt = False):
        """Creates Moran's I Step Output Table.

        INPUTS
        outputTable (str): path to the output table
        displayIt {bool, False}: create output graph?
        """

        #### Allow Overwrite Output ####
        ARCPY.env.overwriteOutput = 1

        #### Get Output Table Name With Extension if Appropriate ####
        outputTable, dbf = UTILS.returnTableName(outputTable)

        #### Set Progressor ####
        ARCPY.SetProgressor("default", ARCPY.GetIDMessage(84008))

        #### Delete Table If Exists ####
        UTILS.passiveDelete(outputTable)

        #### Create Table ####
        outPath, outName = OS.path.split(outputTable)
        try:
            DM.CreateTable(outPath, outName)
        except:
            ARCPY.AddIDMessage("ERROR", 541)
            raise ERROR.ScriptError()

        #### Add Result Fields ####
        validNames = []
        for field in self.outputFields:
            try:
                fieldOut = ARCPY.ValidateFieldName(field, outPath)
                DM.AddField(outputTable, fieldOut, "DOUBLE")
                validNames.append(fieldOut)
            except:
                ARCPY.AddIDMessage("Error", 852, field, outputTable)

        #### Create Insert Cursor ####
        try:
            insert = ARCPY.InsertCursor(outputTable)
        except:
            ARCPY.AddIDMessage("ERROR", 204)
            raise ERROR.ScriptError()

        #### Add Rows to Output Table ####
        for testIter in range(self.nIncrements):
            row = insert.newRow()
            result = self.giResults[testIter]
            for ind, fieldName in enumerate(validNames):
                value = result[ind]
                row.setValue(fieldName, value)
            insert.insertRow(row)

        #### Clean Up ####   
        del insert

        #### Make Table Visable in TOC if *.dbf Had To Be Added ####
        if dbf:
            ARCPY.SetParameterAsText(7, outputTable)

        #### Display Results ####
        if displayIt:
            if "WIN" in SYS.platform.upper():
                #### Set Progressor ####
                ARCPY.SetProgressor("default", ARCPY.GetIDMessage(84186))

                #### Get Image Directory ####
                imageDir = OS.path.join(OS.path.dirname(SYS.argv[0]), 
                                        "Images")

                #### Make List of Fields and Set Template File ####
                yFields = [validNames[-2]]
                xFields = [validNames[0]]

                #### Create Data Series String ####
                dataStr = UTILS.createSeriesStr(xFields, yFields, outputTable)

                #### Make Graph ####
                tee = OS.path.join(imageDir, "MoransIGraph.tee")
                DM.MakeGraph(tee, dataStr, "MoransI_Distances")
                ARCPY.SetParameterAsText(9, "MoransI_Distances")

            else:
                ARCPY.AddIDMessage("Warning", 942)

if __name__ == "__main__":
    setupGlobalI_Increment()

