"""
Tool Name:     Global Moran's I
Source Name:   GlobalI.py
Version:       ArcGIS 10.0
Author:        Environmental Systems Research Institute Inc.
Description:   Computes Global Moran's I statistic
"""

################### Imports ########################
import sys as SYS
import os as OS
import numpy as NUM
import xml.etree.ElementTree as ET
import arcpy as ARCPY
import ErrorUtils as ERROR
import SSUtilities as UTILS
import SSDataObject as SSDO
import Stats as STATS
import WeightsUtilities as WU
import gapy as GAPY
import SSReport as REPORT
import locale as LOCALE
LOCALE.setlocale(LOCALE.LC_ALL, '')

class GlobalI_Step(object):
    """Calculates Global Morans I, Step-Wise Version (Array Based):
    """

    def __init__(self, ssdo, y, weightsFile):

        #### Set Initial Attributes ####
        UTILS.assignClassAttr(self, locals())

        #### Initialize Data ####
        self.initialize()

        #### Construct Based on SWM File or On The Fly ####
        self.construct()

        #### Calculate Moran's I ####
        self.calculate()

    def initialize(self):
        """Populates the instance of the Spatial Statistics Data 
        Object (SSDataObject) and resolves a default distance threshold
        if none given.
        """

        #### Shorthand Attributes ####
        ssdo = self.ssdo
        weightsFile = self.weightsFile
        self.master2Order = ssdo.master2Order

    def construct(self):
        """Constructs the neighborhood structure for each feature and
        dispatches the appropriate values for the calculation of the
        statistic."""

        #### Shorthand Attributes ####
        ssdo = self.ssdo
        numObs = len(self.y)
        master2Order = self.master2Order
        weightsFile = self.weightsFile

        yVar = NUM.var(self.y)
        if NUM.isnan(yVar) or yVar <= 0.0:
            ARCPY.AddIDMessage("Error", 906)
            raise ERROR.ScriptError()

        #### Create Deviation Variables ####
        self.yBar = NUM.mean(self.y)
        self.yDev = self.y - self.yBar

        #### Create Base Data Structures/Variables #### 
        self.numer = 0.0
        self.denom = NUM.sum(self.yDev**2.0)
        self.rowSum = NUM.zeros(numObs)
        self.colSum = NUM.zeros(numObs)
        self.s0 = 0
        self.s1 = 0
        self.wij = {}

        #### Keep Track of Features with No Neighbors ####
        self.idsNoNeighs = []

        #### Open Spatial Weights and Obtain Chars ####
        header = WU.readWeightsHeader(weightsFile)
        f, masterField, spatialRefName, N, rowStandard = header

        #### Check to Assure Complete Set of Weights ####
        if numObs > N:
            ARCPY.AddIDMessage("Error", 842, numObs, N)
            raise ERROR.ScriptError()
        
        #### Check if Selection Set ####
        isSubSet = False
        if numObs < N:
            isSubSet = True
        iterVals = xrange(N)

        #### Calculation For Each Feature ####
        for i in iterVals:
            info = WU.readWeightsEntry(f)
            masterID = info[0]
            if master2Order.has_key(masterID):
                rowInfo = WU.getWeightsValuesSWM(info, master2Order,
                                                 self.yDev, 
                                                 rowStandard = rowStandard,
                                                 isSubSet = isSubSet)
                includeIt = True
            else:
                includeIt = False


            #### Subset Boolean for SWM File ####
            if includeIt:
                #### Parse Row Info ####
                orderID, yiDev, nhIDs, nhVals, weights = rowInfo

                #### Assure Neighbors Exist After Selection ####
                nn = len(nhVals)

                if nn:
                    #### Process Feature Contribution to Moran's I ####
                    self.processRow(orderID, yiDev, nhIDs, 
                                          nhVals, weights) 

                else:
                    #### No Neighbors ####
                    self.addNoNeighbor(masterID, orderID)

        #### Report if All Features Have No Neighbors ####
        if self.s0 == 0:
            ARCPY.AddIDMessage("Error", 908)
            raise ERROR.ScriptError()

        #### Report on Features with No Neighbors ####
        noNeighs = len(self.idsNoNeighs)
        if noNeighs:
            self.idsNoNeighs.sort()
            if noNeighs > 30:
                printNoNeighs = self.idsNoNeighs[0:30]
            else:
                printNoNeighs = self.idsNoNeighs
            ERROR.warningNoNeighbors(numObs, noNeighs, printNoNeighs,
                                     masterField)

    def processRow(self, orderID, yiDev, nhIDs, nhVals, weights):
        """Processes a features contribution to the Moran's I statistic.
        
        INPUTS:
        orderID (int): order in corresponding numpy value arrays
        yiVal (float): value for given feature
        nhIDs (array, nn): neighbor order in corresponding numpy value arrays
        nhVals (array, nn): values for neighboring features (1)
        weights (array, nn): weight values for neighboring features (1)

        NOTES:
        (1)  nn is equal to the number of neighboring features
        """

        #### Numerator Calculation ####
        sumW = weights.sum()
        self.s0 += sumW
        self.numer += NUM.sum(nhVals * weights) * yiDev

        #### Weights Charactersitics Update ####
        c = 0
        for neighID in nhIDs:
            ij = (orderID, neighID)
            ji = (neighID, orderID)
            w = weights[c] 
            self.s1 += w**2.0
            try:
                self.s1 += 2.0 * w * self.wij.pop(ji)
            except:
                self.wij[ij] = w
            self.rowSum[orderID] += w
            self.colSum[neighID] += w
            c += 1

    def calculate(self):
        """Calculate Moran's I Statistic."""

        s0 = self.s0
        s1 = self.s1
        n = len(self.rowSum) * 1.0
        s2 = NUM.sum( (self.rowSum + self.colSum)**2.0 )
        self.s2 = s2
        self.ei = -1. / (n - 1)
        self.squareExpectedI = self.ei**2
        self.n = n
        scale = n / s0
        s02 = s0 * s0
        n2 = n * n
        n2s1 = n2 * s1
        ns2 = n * s2
        self.gi = (scale * (self.numer/self.denom))
        yDev4Sum = NUM.sum(self.yDev**4) / n
        yDevsqsq = (self.denom / n)**2
        b2 = yDev4Sum / yDevsqsq
        self.b2 = b2
        left = n * ((n2 - (3*n) + 3) * s1 - (n*s2) + 3 * (s02))
        right = b2 * ((n2 - n) * s1 - (2*n*s2) + 6 * (s02))
        denom = (n-1) * (n-2) * (n-3) * s02
        num = (left - right) / denom
        self.expectedSquaredI = num 
        self.vi = self.expectedSquaredI - self.squareExpectedI

        #### Assure that Variance is Larger than Zero ####
        if NUM.isnan(self.vi) or self.vi <= 0.0:
            ARCPY.AddIDMessage("Error", 906)
            raise ERROR.ScriptError()

        self.standDev = NUM.sqrt(self.vi)
        self.zi = (self.gi - self.ei)/self.standDev
        self.pVal = STATS.zProb(self.zi, type = 2)

    def report(self, fileName = None):
        """Reports the Moran's I results as a message or to a file.  If
        self.displayIt is set to True, then an html graphical report is
        generated to your default temp directory.

        INPUTS:
        fileName {str, None}: path to a text file to populate with results.
        """

        header = ARCPY.GetIDMessage(84160)
        giString = LOCALE.format("%0.6f", self.gi)
        eiString = LOCALE.format("%0.6f", self.ei)
        viString = LOCALE.format("%0.6f", self.vi) 
        ziString = LOCALE.format("%0.6f", self.zi) 
        pvString = LOCALE.format("%0.6f", self.pVal) 
        row1 = [ARCPY.GetIDMessage(84148), giString]
        row2 = [ARCPY.GetIDMessage(84149), eiString]
        row3 = [ARCPY.GetIDMessage(84150), viString]
        row4 = [ARCPY.GetIDMessage(84151), ziString]
        row5 = [ARCPY.GetIDMessage(84152), pvString]
        results =  [row1, row2, row3, row4, row5]
        outputTable = UTILS.outputTextTable(results, header = header)
        if fileName:
            f = open(fileName, "w")
            f.write(outputTable)
            f.close()
        else:
            ARCPY.AddMessage(outputTable)

        #### Set Formatted Floats ####
        self.giString = giString
        self.eiString = eiString
        self.viString = viString
        self.ziString = ziString
        self.pvString = pvString
        
        #### HTML Report ####
        if self.displayIt:
            self.reportHTML()

        return giString, ziString, pvString

    def reportHTML(self):
        """Generates a graphical html report for Moran's I."""

        #### Shorthand Attributes ####
        zi = self.zi

        #### Progress and Create HTML File Name ####
        writeMSG = ARCPY.GetIDMessage(84228)
        ARCPY.SetProgressor("default", writeMSG)
        ARCPY.AddMessage(writeMSG)
        prefix = ARCPY.GetIDMessage(84227)
        outputDir = UTILS.returnScratchWorkSpace()
        baseDir = UTILS.getBaseFolder(outputDir)
        htmlFile = UTILS.returnScratchName(prefix, fileType = "TEXT", 
                                           scratchWS = baseDir,
                                           extension = "html")

        #### Obtain Correct Images ####
        imageDir = UTILS.getImageDir()
        clustStr = ARCPY.GetIDMessage(84243)
        dispStr = ARCPY.GetIDMessage(84244)
        if zi <= -2.58:
            imageFile = OS.path.join(imageDir, "dispersedValues01.png")
            info = ("1%", dispStr)
            imageBox = OS.path.join(imageDir, "dispersedBox01.png")
        elif (-2.58 < zi <= -1.96):
            imageFile = OS.path.join(imageDir, "dispersedValues05.png")
            info = ("5%", dispStr)
            imageBox = OS.path.join(imageDir, "dispersedBox05.png")
        elif (-1.96 < zi <= -1.65):
            imageFile = OS.path.join(imageDir, "dispersedValues10.png")
            info = ("10%", dispStr)
            imageBox = OS.path.join(imageDir, "dispersedBox10.png")
        elif (-1.65 < zi < 1.65):
            imageFile = OS.path.join(imageDir, "randomValues.png")
            imageBox = OS.path.join(imageDir, "randomBox.png")
        elif (1.65 <= zi < 1.96):
            imageFile = OS.path.join(imageDir, "clusteredValues10.png")
            info = ("10%", clustStr)
            imageBox = OS.path.join(imageDir, "clusteredBox10.png")
        elif (1.96 <= zi < 2.58):
            imageFile = OS.path.join(imageDir, "clusteredValues05.png")
            info = ("5%", clustStr)
            imageBox = OS.path.join(imageDir, "clusteredBox05.png")
        else:
            imageFile = OS.path.join(imageDir, "clusteredValues01.png")
            info = ("1%", clustStr)
            imageBox = OS.path.join(imageDir, "clusteredBox01.png")

        #### Footnote ####
        footStart = ARCPY.GetIDMessage(84230) % zi
        if abs(zi) >= 1.65:
            footEnd = ARCPY.GetIDMessage(84231)
            footerText = footStart + (footEnd % info)
        else:
            footEnd = ARCPY.GetIDMessage(84232)
            footerText = footStart + footEnd

        #### Root Element ####
        title = ARCPY.GetIDMessage(84229)
        reportElement, reportTree = REPORT.xmlReport(title = title)

        #### Begin Graphic SubElement ####
        graphicElement = REPORT.xmlGraphic(reportElement, imageFile, 
                                           footerText = footerText)

        #### Floating Table ####
        rowVals = [ [ARCPY.GetIDMessage(84148), self.giString, ""],
                    [ARCPY.GetIDMessage(84151), self.ziString, imageBox],
                    [ARCPY.GetIDMessage(84152), self.pvString, ""] ]

        fTable = REPORT.xmlTable(graphicElement, rowVals, 
                                 tType = "ssFloat")

        #### Moran Table ####
        rowVals = [ [ARCPY.GetIDMessage(84148), self.giString],
                    [ARCPY.GetIDMessage(84149), self.eiString],
                    [ARCPY.GetIDMessage(84150), self.viString],
                    [ARCPY.GetIDMessage(84151), self.ziString],
                    [ARCPY.GetIDMessage(84152), self.pvString] ]

        mTable = REPORT.xmlTable(reportElement, rowVals,
                                 title = ARCPY.GetIDMessage(84160))

        #### Dataset Table ####
        rowVals = [ [UTILS.addColon(ARCPY.GetIDMessage(84233)), 
                     self.ssdo.inputFC],
                    [UTILS.addColon(ARCPY.GetIDMessage(84016)), 
                     self.varName],
                    [UTILS.addColon(ARCPY.GetIDMessage(84234)), 
                     WU.wTypeDispatch[self.wType]],
                    [UTILS.addColon(ARCPY.GetIDMessage(84235)),
                     self.concept],
                    [UTILS.addColon(ARCPY.GetIDMessage(84236)), 
                     str(self.rowStandard)],
                    [UTILS.addColon(ARCPY.GetIDMessage(84237)), 
                     str(self.threshold)],
                    [UTILS.addColon(ARCPY.GetIDMessage(84238)), 
                     str(self.weightsFile)] ]

        dTable = REPORT.xmlTable(reportElement, rowVals,
                                 title = ARCPY.GetIDMessage(84239))

        #### Create HTML ####
        html = REPORT.report2html(reportTree, htmlFile)
        ARCPY.AddMessage(htmlFile)
        ARCPY.SetParameterAsText(11, htmlFile)

    def addNoNeighbor(self, masterID, orderID):
        """Accounting for features with no neighbors.
        
        INPUTS:
        masterID (int): unique ID value that has no neighbors
        orderID (int): order in corresponding numpy value arrays
        """

        self.idsNoNeighs.append(masterID)

if __name__ == "__main__":
    setupGlobalI()

